@@ -226,8 +226,8 @@ pub(crate) fn invalid_first_argument_name(checker: &Checker, scope: &Scope) {
         decorator_list,
         parent_scope,
         semantic,
-        &checker.settings.pep8_naming.classmethod_decorators,
-        &checker.settings.pep8_naming.staticmethod_decorators,
+        &checker.settings().pep8_naming.classmethod_decorators,
+        &checker.settings().pep8_naming.staticmethod_decorators,
     ) {
         function_type::FunctionType::Function | function_type::FunctionType::StaticMethod => {
             return;
@@ -243,7 +243,7 @@ pub(crate) fn invalid_first_argument_name(checker: &Checker, scope: &Scope) {
             return;
         }
     };
-    if !checker.enabled(function_type.rule()) {
+    if !checker.is_rule_enabled(function_type.rule()) {
         return;
     }
 
@@ -260,7 +260,7 @@ pub(crate) fn invalid_first_argument_name(checker: &Checker, scope: &Scope) {
 
     if &self_or_cls.name == function_type.valid_first_argument_name()
         || checker
-            .settings
+            .settings()
             .pep8_naming
             .ignore_names
             .matches(&self_or_cls.name)
