@@ -1,9 +1,10 @@
 use ruff_python_ast::Identifier;
 
-use ruff_diagnostics::{Diagnostic, Violation};
-use ruff_macros::{derive_message_formats, violation};
+use ruff_macros::{ViolationMetadata, derive_message_formats};
 use ruff_text_size::Ranged;
 
+use crate::Violation;
+use crate::checkers::ast::Checker;
 use crate::rules::pycodestyle::helpers::is_ambiguous_name;
 
 /// ## What it does
@@ -14,18 +15,18 @@ use crate::rules::pycodestyle::helpers::is_ambiguous_name;
 /// numerals one and zero. When tempted to use 'l', use 'L' instead.
 ///
 /// ## Example
+///
 /// ```python
-/// def l(x):
-///     ...
+/// def l(x): ...
 /// ```
 ///
 /// Use instead:
+///
 /// ```python
-/// def long_name(x):
-///     ...
+/// def long_name(x): ...
 /// ```
-#[violation]
-pub struct AmbiguousFunctionName(pub String);
+#[derive(ViolationMetadata)]
+pub(crate) struct AmbiguousFunctionName(pub String);
 
 impl Violation for AmbiguousFunctionName {
     #[derive_message_formats]
@@ -36,13 +37,8 @@ impl Violation for AmbiguousFunctionName {
 }
 
 /// E743
-pub(crate) fn ambiguous_function_name(name: &Identifier) -> Option<Diagnostic> {
+pub(crate) fn ambiguous_function_name(checker: &Checker, name: &Identifier) {
     if is_ambiguous_name(name) {
-        Some(Diagnostic::new(
-            AmbiguousFunctionName(name.to_string()),
-            name.range(),
-        ))
-    } else {
-        None
+        checker.report_diagnostic(AmbiguousFunctionName(name.to_string()), name.range());
     }
 }
