@@ -1,11 +1,11 @@
 use ruff_python_ast::{self as ast, Arguments, Expr, Keyword};
 use ruff_text_size::{Ranged, TextRange};
 
-use ruff_diagnostics::{Diagnostic, Edit, Fix, FixAvailability, Violation};
-use ruff_macros::{derive_message_formats, violation};
+use ruff_macros::{ViolationMetadata, derive_message_formats};
 use ruff_python_semantic::SemanticModel;
 
 use crate::checkers::ast::Checker;
+use crate::{Edit, Fix, FixAvailability, Violation};
 
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub(crate) enum MinMax {
@@ -21,6 +21,7 @@ pub(crate) enum MinMax {
 /// readability.
 ///
 /// ## Example
+///
 /// ```python
 /// minimum = min(1, 2, min(3, 4, 5))
 /// maximum = max(1, 2, max(3, 4, 5))
@@ -28,17 +29,53 @@ pub(crate) enum MinMax {
 /// ```
 ///
 /// Use instead:
+///
 /// ```python
 /// minimum = min(1, 2, 3, 4, 5)
 /// maximum = max(1, 2, 3, 4, 5)
 /// diff = maximum - minimum
 /// ```
 ///
+/// ## Known issues
+///
+/// The resulting code may be slower and use more memory, especially for nested iterables. For
+/// example, this code:
+///
+/// ```python
+/// iterable = range(3)
+/// min(1, min(iterable))
+/// ```
+///
+/// will be fixed to:
+///
+/// ```python
+/// iterable = range(3)
+/// min(1, *iterable)
+/// ```
+///
+/// At least on current versions of CPython, this allocates a collection for the whole iterable
+/// before calling `min` and could cause performance regressions, at least for large iterables.
+///
+/// ## Fix safety
+///
+/// This fix is always unsafe and may change the program's behavior for types without full
+/// equivalence relations, such as float comparisons involving `NaN`.
+///
+/// ```python
+/// print(min(2.0, min(float("nan"), 1.0)))  # before fix: 2.0
+/// print(min(2.0, float("nan"), 1.0))  # after fix: 1.0
+///
+/// print(max(1.0, max(float("nan"), 2.0)))  # before fix: 1.0
+/// print(max(1.0, float("nan"), 2.0))  # after fix: 2.0
+/// ```
+///
+/// The fix will also remove any comments within the outer call.
+///
 /// ## References
 /// - [Python documentation: `min`](https://docs.python.org/3/library/functions.html#min)
 /// - [Python documentation: `max`](https://docs.python.org/3/library/functions.html#max)
-#[violation]
-pub struct NestedMinMax {
+#[derive(ViolationMetadata)]
+pub(crate) struct NestedMinMax {
     func: MinMax,
 }
 
@@ -97,22 +134,25 @@ fn collect_nested_args(min_max: MinMax, args: &[Expr], semantic: &SemanticModel)
                         args,
                         keywords,
                         range: _,
+                        node_index: _,
                     },
                 range: _,
+                node_index: _,
             }) = arg
             {
-                if let [arg] = &**args {
-                    if arg.as_starred_expr().is_none() {
-                        let new_arg = Expr::Starred(ast::ExprStarred {
-                            value: Box::new(arg.clone()),
-                            ctx: ast::ExprContext::Load,
-                            range: TextRange::default(),
-                        });
-                        new_args.push(new_arg);
-                        continue;
-                    }
-                }
                 if MinMax::try_from_call(func, keywords, semantic) == Some(min_max) {
+                    if let [arg] = &**args {
+                        if arg.as_starred_expr().is_none() {
+                            let new_arg = Expr::Starred(ast::ExprStarred {
+                                value: Box::new(arg.clone()),
+                                ctx: ast::ExprContext::Load,
+                                range: TextRange::default(),
+                                node_index: ruff_python_ast::AtomicNodeIndex::dummy(),
+                            });
+                            new_args.push(new_arg);
+                            continue;
+                        }
+                    }
                     inner(min_max, args, semantic, new_args);
                     continue;
                 }
@@ -128,7 +168,7 @@ fn collect_nested_args(min_max: MinMax, args: &[Expr], semantic: &SemanticModel)
 
 /// PLW3301
 pub(crate) fn nested_min_max(
-    checker: &mut Checker,
+    checker: &Checker,
     expr: &Expr,
     func: &Expr,
     args: &[Expr],
@@ -137,9 +177,11 @@ pub(crate) fn nested_min_max(
     let Some(min_max) = MinMax::try_from_call(func, keywords, checker.semantic()) else {
         return;
     };
-
-    if matches!(&args, [Expr::Call(ast::ExprCall { arguments: Arguments {args, .. }, .. })] if args.len() == 1)
-    {
+    // It's only safe to flatten nested calls if the outer call has more than one argument.
+    // When the outer call has a single argument, flattening would change the semantics by
+    // changing the shape of the call from treating the inner result as an iterable (or a scalar)
+    // to passing multiple arguments directly, which can lead to behavioral changes.
+    if args.len() < 2 {
         return;
     }
 
@@ -154,25 +196,22 @@ pub(crate) fn nested_min_max(
         };
         MinMax::try_from_call(func.as_ref(), keywords.as_ref(), checker.semantic()) == Some(min_max)
     }) {
-        let mut diagnostic = Diagnostic::new(NestedMinMax { func: min_max }, expr.range());
-        if !checker
-            .comment_ranges()
-            .has_comments(expr, checker.locator())
-        {
-            let flattened_expr = Expr::Call(ast::ExprCall {
-                func: Box::new(func.clone()),
-                arguments: Arguments {
-                    args: collect_nested_args(min_max, args, checker.semantic()).into_boxed_slice(),
-                    keywords: Box::from(keywords),
-                    range: TextRange::default(),
-                },
+        let mut diagnostic =
+            checker.report_diagnostic(NestedMinMax { func: min_max }, expr.range());
+        let flattened_expr = Expr::Call(ast::ExprCall {
+            func: Box::new(func.clone()),
+            arguments: Arguments {
+                args: collect_nested_args(min_max, args, checker.semantic()).into_boxed_slice(),
+                keywords: Box::from(keywords),
                 range: TextRange::default(),
-            });
-            diagnostic.set_fix(Fix::unsafe_edit(Edit::range_replacement(
-                checker.generator().expr(&flattened_expr),
-                expr.range(),
-            )));
-        }
-        checker.diagnostics.push(diagnostic);
+                node_index: ruff_python_ast::AtomicNodeIndex::dummy(),
+            },
+            range: TextRange::default(),
+            node_index: ruff_python_ast::AtomicNodeIndex::dummy(),
+        });
+        diagnostic.set_fix(Fix::unsafe_edit(Edit::range_replacement(
+            checker.generator().expr(&flattened_expr),
+            expr.range(),
+        )));
     }
 }
