@@ -85,6 +85,13 @@ pub struct InlayHintSettings {
     /// foo("x="1)
     /// ```
     pub call_argument_names: bool,
+    // Add any new setting that enables additional inlays to `any_enabled`.
+}
+
+impl InlayHintSettings {
+    pub fn any_enabled(&self) -> bool {
+        self.variable_types || self.call_argument_names
+    }
 }
 
 impl Default for InlayHintSettings {
