@@ -7,16 +7,19 @@ use crate::suppression::{INVALID_IGNORE_COMMENT, UNKNOWN_RULE, UNUSED_IGNORE_COM
 pub use db::Db;
 pub use module_name::ModuleName;
 pub use module_resolver::{
-    Module, SearchPathValidationError, SearchPaths, resolve_module, resolve_real_module,
-    system_module_search_paths,
+    Module, SearchPath, SearchPathValidationError, SearchPaths, list_modules, resolve_module,
+    resolve_real_module, system_module_search_paths,
 };
 pub use program::{
     Program, ProgramSettings, PythonVersionFileSource, PythonVersionSource,
     PythonVersionWithSource, SearchPathSettings,
 };
 pub use python_platform::PythonPlatform;
-pub use semantic_model::{Completion, CompletionKind, HasType, NameKind, SemanticModel};
+pub use semantic_model::{
+    Completion, CompletionKind, HasDefinition, HasType, NameKind, SemanticModel,
+};
 pub use site_packages::{PythonEnvironment, SitePackagesPaths, SysPrefixPathOrigin};
+pub use types::DisplaySettings;
 pub use types::ide_support::{
     ImportAliasResolution, ResolvedDefinition, definitions_for_attribute,
     definitions_for_imported_symbol, definitions_for_name, map_stub_definition,
