@@ -24,9 +24,11 @@ use itertools::{Either, EitherOrBoth, Itertools};
 use crate::semantic_index::definition::Definition;
 use crate::types::Truthiness;
 use crate::types::class::{ClassType, KnownClass};
+use crate::types::constraints::{Constraints, IteratorConstraintsExtension};
 use crate::types::{
     ApplyTypeMappingVisitor, BoundTypeVarInstance, HasRelationToVisitor, IsDisjointVisitor,
-    NormalizedVisitor, Type, TypeMapping, TypeRelation, TypeVarVariance, UnionBuilder, UnionType,
+    IsEquivalentVisitor, NormalizedVisitor, Type, TypeMapping, TypeRelation, TypeVarVariance,
+    UnionBuilder, UnionType,
 };
 use crate::util::subscript::{Nth, OutOfBoundsError, PyIndex, PySlice, StepSizeZeroError};
 use crate::{Db, FxOrderSet, Program};
@@ -254,19 +256,25 @@ impl<'db> TupleType<'db> {
             .find_legacy_typevars(db, binding_context, typevars);
     }
 
-    pub(crate) fn has_relation_to_impl(
+    pub(crate) fn has_relation_to_impl<C: Constraints<'db>>(
         self,
         db: &'db dyn Db,
         other: Self,
         relation: TypeRelation,
-        visitor: &HasRelationToVisitor<'db>,
-    ) -> bool {
+        visitor: &HasRelationToVisitor<'db, C>,
+    ) -> C {
         self.tuple(db)
             .has_relation_to_impl(db, other.tuple(db), relation, visitor)
     }
 
-    pub(crate) fn is_equivalent_to(self, db: &'db dyn Db, other: Self) -> bool {
-        self.tuple(db).is_equivalent_to(db, other.tuple(db))
+    pub(crate) fn is_equivalent_to_impl<C: Constraints<'db>>(
+        self,
+        db: &'db dyn Db,
+        other: Self,
+        visitor: &IsEquivalentVisitor<'db, C>,
+    ) -> C {
+        self.tuple(db)
+            .is_equivalent_to_impl(db, other.tuple(db), visitor)
     }
 
     pub(crate) fn is_single_valued(self, db: &'db dyn Db) -> bool {
@@ -409,56 +417,76 @@ impl<'db> FixedLengthTuple<Type<'db>> {
         }
     }
 
-    fn has_relation_to_impl(
+    fn has_relation_to_impl<C: Constraints<'db>>(
         &self,
         db: &'db dyn Db,
         other: &Tuple<Type<'db>>,
         relation: TypeRelation,
-        visitor: &HasRelationToVisitor<'db>,
-    ) -> bool {
+        visitor: &HasRelationToVisitor<'db, C>,
+    ) -> C {
         match other {
-            Tuple::Fixed(other) => {
-                self.0.len() == other.0.len()
-                    && (self.0.iter()).zip(&other.0).all(|(self_ty, other_ty)| {
-                        self_ty.has_relation_to_impl(db, *other_ty, relation, visitor)
-                    })
-            }
+            Tuple::Fixed(other) => C::from_bool(db, self.0.len() == other.0.len()).and(db, || {
+                (self.0.iter().zip(&other.0)).when_all(db, |(self_ty, other_ty)| {
+                    self_ty.has_relation_to_impl(db, *other_ty, relation, visitor)
+                })
+            }),
 
             Tuple::Variable(other) => {
                 // This tuple must have enough elements to match up with the other tuple's prefix
                 // and suffix, and each of those elements must pairwise satisfy the relation.
+                let mut result = C::always_satisfiable(db);
                 let mut self_iter = self.0.iter();
                 for other_ty in &other.prefix {
                     let Some(self_ty) = self_iter.next() else {
-                        return false;
+                        return C::unsatisfiable(db);
                     };
-                    if !self_ty.has_relation_to_impl(db, *other_ty, relation, visitor) {
-                        return false;
+                    let element_constraints =
+                        self_ty.has_relation_to_impl(db, *other_ty, relation, visitor);
+                    if result
+                        .intersect(db, element_constraints)
+                        .is_never_satisfied(db)
+                    {
+                        return result;
                     }
                 }
                 for other_ty in other.suffix.iter().rev() {
                     let Some(self_ty) = self_iter.next_back() else {
-                        return false;
+                        return C::unsatisfiable(db);
                     };
-                    if !self_ty.has_relation_to_impl(db, *other_ty, relation, visitor) {
-                        return false;
+                    let element_constraints =
+                        self_ty.has_relation_to_impl(db, *other_ty, relation, visitor);
+                    if result
+                        .intersect(db, element_constraints)
+                        .is_never_satisfied(db)
+                    {
+                        return result;
                     }
                 }
 
                 // In addition, any remaining elements in this tuple must satisfy the
                 // variable-length portion of the other tuple.
-                self_iter.all(|self_ty| {
-                    self_ty.has_relation_to_impl(db, other.variable, relation, visitor)
+                result.and(db, || {
+                    self_iter.when_all(db, |self_ty| {
+                        self_ty.has_relation_to_impl(db, other.variable, relation, visitor)
+                    })
                 })
             }
         }
     }
 
-    fn is_equivalent_to(&self, db: &'db dyn Db, other: &Self) -> bool {
-        self.0.len() == other.0.len()
-            && (self.0.iter())
+    fn is_equivalent_to_impl<C: Constraints<'db>>(
+        &self,
+        db: &'db dyn Db,
+        other: &Self,
+        visitor: &IsEquivalentVisitor<'db, C>,
+    ) -> C {
+        C::from_bool(db, self.0.len() == other.0.len()).and(db, || {
+            (self.0.iter())
                 .zip(&other.0)
-                .all(|(self_ty, other_ty)| self_ty.is_equivalent_to(db, *other_ty))
+                .when_all(db, |(self_ty, other_ty)| {
+                    self_ty.is_equivalent_to_impl(db, *other_ty, visitor)
+                })
+        })
     }
 
     fn is_single_valued(&self, db: &'db dyn Db) -> bool {
@@ -717,13 +745,13 @@ impl<'db> VariableLengthTuple<Type<'db>> {
         }
     }
 
-    fn has_relation_to_impl(
+    fn has_relation_to_impl<C: Constraints<'db>>(
         &self,
         db: &'db dyn Db,
         other: &Tuple<Type<'db>>,
         relation: TypeRelation,
-        visitor: &HasRelationToVisitor<'db>,
-    ) -> bool {
+        visitor: &HasRelationToVisitor<'db, C>,
+    ) -> C {
         match other {
             Tuple::Fixed(other) => {
                 // The `...` length specifier of a variable-length tuple type is interpreted
@@ -738,32 +766,43 @@ impl<'db> VariableLengthTuple<Type<'db>> {
                 // length.
                 if relation == TypeRelation::Subtyping || !matches!(self.variable, Type::Dynamic(_))
                 {
-                    return false;
+                    return C::unsatisfiable(db);
                 }
 
                 // In addition, the other tuple must have enough elements to match up with this
                 // tuple's prefix and suffix, and each of those elements must pairwise satisfy the
                 // relation.
+                let mut result = C::always_satisfiable(db);
                 let mut other_iter = other.elements().copied();
                 for self_ty in self.prenormalized_prefix_elements(db, None) {
                     let Some(other_ty) = other_iter.next() else {
-                        return false;
+                        return C::unsatisfiable(db);
                     };
-                    if !self_ty.has_relation_to_impl(db, other_ty, relation, visitor) {
-                        return false;
+                    let element_constraints =
+                        self_ty.has_relation_to_impl(db, other_ty, relation, visitor);
+                    if result
+                        .intersect(db, element_constraints)
+                        .is_never_satisfied(db)
+                    {
+                        return result;
                     }
                 }
                 let suffix: Vec<_> = self.prenormalized_suffix_elements(db, None).collect();
                 for self_ty in suffix.iter().rev() {
                     let Some(other_ty) = other_iter.next_back() else {
-                        return false;
+                        return C::unsatisfiable(db);
                     };
-                    if !self_ty.has_relation_to_impl(db, other_ty, relation, visitor) {
-                        return false;
+                    let element_constraints =
+                        self_ty.has_relation_to_impl(db, other_ty, relation, visitor);
+                    if result
+                        .intersect(db, element_constraints)
+                        .is_never_satisfied(db)
+                    {
+                        return result;
                     }
                 }
 
-                true
+                result
             }
 
             Tuple::Variable(other) => {
@@ -781,12 +820,13 @@ impl<'db> VariableLengthTuple<Type<'db>> {
                 // The overlapping parts of the prefixes and suffixes must satisfy the relation.
                 // Any remaining parts must satisfy the relation with the other tuple's
                 // variable-length part.
-                if !self
-                    .prenormalized_prefix_elements(db, self_prenormalize_variable)
+                let mut result = C::always_satisfiable(db);
+                let pairwise = (self.prenormalized_prefix_elements(db, self_prenormalize_variable))
                     .zip_longest(
                         other.prenormalized_prefix_elements(db, other_prenormalize_variable),
-                    )
-                    .all(|pair| match pair {
+                    );
+                for pair in pairwise {
+                    let pair_constraints = match pair {
                         EitherOrBoth::Both(self_ty, other_ty) => {
                             self_ty.has_relation_to_impl(db, other_ty, relation, visitor)
                         }
@@ -796,11 +836,15 @@ impl<'db> VariableLengthTuple<Type<'db>> {
                         EitherOrBoth::Right(_) => {
                             // The rhs has a required element that the lhs is not guaranteed to
                             // provide.
-                            false
+                            return C::unsatisfiable(db);
                         }
-                    })
-                {
-                    return false;
+                    };
+                    if result
+                        .intersect(db, pair_constraints)
+                        .is_never_satisfied(db)
+                    {
+                        return result;
+                    }
                 }
 
                 let self_suffix: Vec<_> = self
@@ -809,9 +853,9 @@ impl<'db> VariableLengthTuple<Type<'db>> {
                 let other_suffix: Vec<_> = other
                     .prenormalized_suffix_elements(db, other_prenormalize_variable)
                     .collect();
-                if !(self_suffix.iter().rev())
-                    .zip_longest(other_suffix.iter().rev())
-                    .all(|pair| match pair {
+                let pairwise = (self_suffix.iter().rev()).zip_longest(other_suffix.iter().rev());
+                for pair in pairwise {
+                    let pair_constraints = match pair {
                         EitherOrBoth::Both(self_ty, other_ty) => {
                             self_ty.has_relation_to_impl(db, *other_ty, relation, visitor)
                         }
@@ -821,34 +865,54 @@ impl<'db> VariableLengthTuple<Type<'db>> {
                         EitherOrBoth::Right(_) => {
                             // The rhs has a required element that the lhs is not guaranteed to
                             // provide.
-                            false
+                            return C::unsatisfiable(db);
                         }
-                    })
-                {
-                    return false;
+                    };
+                    if result
+                        .intersect(db, pair_constraints)
+                        .is_never_satisfied(db)
+                    {
+                        return result;
+                    }
                 }
 
                 // And lastly, the variable-length portions must satisfy the relation.
-                self.variable
-                    .has_relation_to_impl(db, other.variable, relation, visitor)
+                result.and(db, || {
+                    self.variable
+                        .has_relation_to_impl(db, other.variable, relation, visitor)
+                })
             }
         }
     }
 
-    fn is_equivalent_to(&self, db: &'db dyn Db, other: &Self) -> bool {
-        self.variable.is_equivalent_to(db, other.variable)
-            && (self.prenormalized_prefix_elements(db, None))
-                .zip_longest(other.prenormalized_prefix_elements(db, None))
-                .all(|pair| match pair {
-                    EitherOrBoth::Both(self_ty, other_ty) => self_ty.is_equivalent_to(db, other_ty),
-                    EitherOrBoth::Left(_) | EitherOrBoth::Right(_) => false,
-                })
-            && (self.prenormalized_suffix_elements(db, None))
-                .zip_longest(other.prenormalized_suffix_elements(db, None))
-                .all(|pair| match pair {
-                    EitherOrBoth::Both(self_ty, other_ty) => self_ty.is_equivalent_to(db, other_ty),
-                    EitherOrBoth::Left(_) | EitherOrBoth::Right(_) => false,
-                })
+    fn is_equivalent_to_impl<C: Constraints<'db>>(
+        &self,
+        db: &'db dyn Db,
+        other: &Self,
+        visitor: &IsEquivalentVisitor<'db, C>,
+    ) -> C {
+        self.variable
+            .is_equivalent_to_impl(db, other.variable, visitor)
+            .and(db, || {
+                (self.prenormalized_prefix_elements(db, None))
+                    .zip_longest(other.prenormalized_prefix_elements(db, None))
+                    .when_all(db, |pair| match pair {
+                        EitherOrBoth::Both(self_ty, other_ty) => {
+                            self_ty.is_equivalent_to_impl(db, other_ty, visitor)
+                        }
+                        EitherOrBoth::Left(_) | EitherOrBoth::Right(_) => C::unsatisfiable(db),
+                    })
+            })
+            .and(db, || {
+                (self.prenormalized_suffix_elements(db, None))
+                    .zip_longest(other.prenormalized_suffix_elements(db, None))
+                    .when_all(db, |pair| match pair {
+                        EitherOrBoth::Both(self_ty, other_ty) => {
+                            self_ty.is_equivalent_to_impl(db, other_ty, visitor)
+                        }
+                        EitherOrBoth::Left(_) | EitherOrBoth::Right(_) => C::unsatisfiable(db),
+                    })
+            })
     }
 }
 
@@ -1027,13 +1091,13 @@ impl<'db> Tuple<Type<'db>> {
         }
     }
 
-    fn has_relation_to_impl(
+    fn has_relation_to_impl<C: Constraints<'db>>(
         &self,
         db: &'db dyn Db,
         other: &Self,
         relation: TypeRelation,
-        visitor: &HasRelationToVisitor<'db>,
-    ) -> bool {
+        visitor: &HasRelationToVisitor<'db, C>,
+    ) -> C {
         match self {
             Tuple::Fixed(self_tuple) => {
                 self_tuple.has_relation_to_impl(db, other, relation, visitor)
@@ -1044,96 +1108,95 @@ impl<'db> Tuple<Type<'db>> {
         }
     }
 
-    fn is_equivalent_to(&self, db: &'db dyn Db, other: &Self) -> bool {
-        match (self, other) {
-            (Tuple::Fixed(self_tuple), Tuple::Fixed(other_tuple)) => {
-                self_tuple.is_equivalent_to(db, other_tuple)
-            }
-            (Tuple::Variable(self_tuple), Tuple::Variable(other_tuple)) => {
-                self_tuple.is_equivalent_to(db, other_tuple)
-            }
-            (Tuple::Fixed(_), Tuple::Variable(_)) | (Tuple::Variable(_), Tuple::Fixed(_)) => false,
-        }
-    }
-
-    pub(super) fn is_disjoint_from_impl(
+    fn is_equivalent_to_impl<C: Constraints<'db>>(
         &self,
         db: &'db dyn Db,
         other: &Self,
-        visitor: &IsDisjointVisitor<'db>,
-    ) -> bool {
+        visitor: &IsEquivalentVisitor<'db, C>,
+    ) -> C {
+        match (self, other) {
+            (Tuple::Fixed(self_tuple), Tuple::Fixed(other_tuple)) => {
+                self_tuple.is_equivalent_to_impl(db, other_tuple, visitor)
+            }
+            (Tuple::Variable(self_tuple), Tuple::Variable(other_tuple)) => {
+                self_tuple.is_equivalent_to_impl(db, other_tuple, visitor)
+            }
+            (Tuple::Fixed(_), Tuple::Variable(_)) | (Tuple::Variable(_), Tuple::Fixed(_)) => {
+                C::unsatisfiable(db)
+            }
+        }
+    }
+
+    pub(super) fn is_disjoint_from_impl<C: Constraints<'db>>(
+        &self,
+        db: &'db dyn Db,
+        other: &Self,
+        visitor: &IsDisjointVisitor<'db, C>,
+    ) -> C {
         // Two tuples with an incompatible number of required elements must always be disjoint.
         let (self_min, self_max) = self.len().size_hint();
         let (other_min, other_max) = other.len().size_hint();
         if self_max.is_some_and(|max| max < other_min) {
-            return true;
+            return C::always_satisfiable(db);
         }
         if other_max.is_some_and(|max| max < self_min) {
-            return true;
+            return C::always_satisfiable(db);
         }
 
         // If any of the required elements are pairwise disjoint, the tuples are disjoint as well.
         #[allow(clippy::items_after_statements)]
-        fn any_disjoint<'s, 'db>(
+        fn any_disjoint<'s, 'db, C: Constraints<'db>>(
             db: &'db dyn Db,
             a: impl IntoIterator<Item = &'s Type<'db>>,
             b: impl IntoIterator<Item = &'s Type<'db>>,
-            visitor: &IsDisjointVisitor<'db>,
-        ) -> bool
+            visitor: &IsDisjointVisitor<'db, C>,
+        ) -> C
         where
             'db: 's,
         {
-            a.into_iter().zip(b).any(|(self_element, other_element)| {
+            (a.into_iter().zip(b)).when_any(db, |(self_element, other_element)| {
                 self_element.is_disjoint_from_impl(db, *other_element, visitor)
             })
         }
 
         match (self, other) {
             (Tuple::Fixed(self_tuple), Tuple::Fixed(other_tuple)) => {
-                if any_disjoint(db, self_tuple.elements(), other_tuple.elements(), visitor) {
-                    return true;
-                }
+                any_disjoint(db, self_tuple.elements(), other_tuple.elements(), visitor)
             }
 
-            (Tuple::Variable(self_tuple), Tuple::Variable(other_tuple)) => {
-                if any_disjoint(
-                    db,
-                    self_tuple.prefix_elements(),
-                    other_tuple.prefix_elements(),
-                    visitor,
-                ) {
-                    return true;
-                }
-                if any_disjoint(
+            // Note that we don't compare the variable-length portions; two pure homogeneous tuples
+            // `tuple[A, ...]` and `tuple[B, ...]` can never be disjoint even if A and B are
+            // disjoint, because `tuple[()]` would be assignable to both.
+            (Tuple::Variable(self_tuple), Tuple::Variable(other_tuple)) => any_disjoint(
+                db,
+                self_tuple.prefix_elements(),
+                other_tuple.prefix_elements(),
+                visitor,
+            )
+            .or(db, || {
+                any_disjoint(
                     db,
                     self_tuple.suffix_elements().rev(),
                     other_tuple.suffix_elements().rev(),
                     visitor,
-                ) {
-                    return true;
-                }
-            }
+                )
+            }),
 
             (Tuple::Fixed(fixed), Tuple::Variable(variable))
             | (Tuple::Variable(variable), Tuple::Fixed(fixed)) => {
-                if any_disjoint(db, fixed.elements(), variable.prefix_elements(), visitor) {
-                    return true;
-                }
-                if any_disjoint(
+                any_disjoint(db, fixed.elements(), variable.prefix_elements(), visitor).or(
                     db,
-                    fixed.elements().rev(),
-                    variable.suffix_elements().rev(),
-                    visitor,
-                ) {
-                    return true;
-                }
+                    || {
+                        any_disjoint(
+                            db,
+                            fixed.elements().rev(),
+                            variable.suffix_elements().rev(),
+                            visitor,
+                        )
+                    },
+                )
             }
         }
-
-        // Two pure homogeneous tuples `tuple[A, ...]` and `tuple[B, ...]` can never be
-        // disjoint even if A and B are disjoint, because `tuple[()]` would be assignable to
-        // both.
-        false
     }
 
     pub(crate) fn is_single_valued(&self, db: &'db dyn Db) -> bool {
