@@ -24,6 +24,9 @@ import {
   Severity,
   type Workspace,
   CompletionKind,
+  type FileHandle,
+  DocumentHighlight,
+  DocumentHighlightKind,
 } from "ty_wasm";
 import { FileId, ReadonlyFiles } from "../Playground";
 import { isPythonFile } from "./Files";
@@ -42,6 +45,9 @@ type Props = {
   onChange(content: string): void;
   onMount(editor: IStandaloneCodeEditor, monaco: Monaco): void;
   onOpenFile(file: FileId): void;
+  onVendoredFileChange: (vendoredFileHandle: FileHandle) => void;
+  onBackToUserFile: () => void;
+  isViewingVendoredFile: boolean;
 };
 
 export default function Editor({
@@ -55,6 +61,9 @@ export default function Editor({
   onChange,
   onMount,
   onOpenFile,
+  onVendoredFileChange,
+  onBackToUserFile,
+  isViewingVendoredFile = false,
 }: Props) {
   const serverRef = useRef<PlaygroundServer | null>(null);
 
@@ -63,6 +72,8 @@ export default function Editor({
       files,
       workspace,
       onOpenFile,
+      onVendoredFileChange,
+      onBackToUserFile,
     });
   }
 
@@ -79,9 +90,12 @@ export default function Editor({
 
   const handleChange = useCallback(
     (value: string | undefined) => {
-      onChange(value ?? "");
+      // Don't update file content when viewing vendored files
+      if (!isViewingVendoredFile) {
+        onChange(value ?? "");
+      }
     },
-    [onChange],
+    [onChange, isViewingVendoredFile],
   );
 
   useEffect(() => {
@@ -98,10 +112,12 @@ export default function Editor({
     (editor, instance) => {
       serverRef.current?.dispose();
 
-      const server = new PlaygroundServer(instance, {
+      const server = new PlaygroundServer(instance, editor, {
         workspace,
         files,
         onOpenFile,
+        onVendoredFileChange,
+        onBackToUserFile,
       });
 
       server.updateDiagnostics(diagnostics);
@@ -110,7 +126,15 @@ export default function Editor({
       onMount(editor, instance);
     },
 
-    [files, onOpenFile, workspace, onMount, diagnostics],
+    [
+      files,
+      onOpenFile,
+      workspace,
+      onMount,
+      diagnostics,
+      onVendoredFileChange,
+      onBackToUserFile,
+    ],
   );
 
   return (
@@ -119,7 +143,7 @@ export default function Editor({
       onMount={handleMount}
       options={{
         fixedOverflowWidgets: true,
-        readOnly: false,
+        readOnly: isViewingVendoredFile, // Make editor read-only for vendored files
         minimap: { enabled: false },
         fontSize: 14,
         roundedSelection: false,
@@ -141,6 +165,8 @@ interface PlaygroundServerProps {
   workspace: Workspace;
   files: ReadonlyFiles;
   onOpenFile: (file: FileId) => void;
+  onVendoredFileChange: (vendoredFileHandle: FileHandle) => void;
+  onBackToUserFile: () => void;
 }
 
 class PlaygroundServer
@@ -156,7 +182,8 @@ class PlaygroundServer
     languages.CompletionItemProvider,
     languages.DocumentSemanticTokensProvider,
     languages.DocumentRangeSemanticTokensProvider,
-    languages.SignatureHelpProvider
+    languages.SignatureHelpProvider,
+    languages.DocumentHighlightProvider
 {
   private typeDefinitionProviderDisposable: IDisposable;
   private declarationProviderDisposable: IDisposable;
@@ -170,9 +197,19 @@ class PlaygroundServer
   private semanticTokensDisposable: IDisposable;
   private rangeSemanticTokensDisposable: IDisposable;
   private signatureHelpDisposable: IDisposable;
+  private documentHighlightDisposable: IDisposable;
+  // Cache for vendored file handles
+  private vendoredFileHandles = new Map<string, FileHandle>();
+
+  private getVendoredPath(uri: Uri): string {
+    // Monaco parses "vendored://stdlib/typing.pyi" as authority="stdlib", path="/typing.pyi"
+    // We need to reconstruct the full path
+    return uri.authority ? `${uri.authority}${uri.path}` : uri.path;
+  }
 
   constructor(
     private monaco: Monaco,
+    private editor: IStandaloneCodeEditor,
     private props: PlaygroundServerProps,
   ) {
     this.typeDefinitionProviderDisposable =
@@ -207,6 +244,11 @@ class PlaygroundServer
       monaco.languages.registerDocumentFormattingEditProvider("python", this);
     this.signatureHelpDisposable =
       monaco.languages.registerSignatureHelpProvider("python", this);
+    this.documentHighlightDisposable =
+      monaco.languages.registerDocumentHighlightProvider("python", this);
+
+    // Register Esc key command
+    editor.addCommand(monaco.KeyCode.Escape, this.props.onBackToUserFile);
   }
 
   triggerCharacters: string[] = ["."];
@@ -223,19 +265,12 @@ class PlaygroundServer
   provideDocumentSemanticTokens(
     model: editor.ITextModel,
   ): languages.SemanticTokens | null {
-    const selectedFile = this.props.files.selected;
-
-    if (selectedFile == null) {
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
       return null;
     }
 
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
-      return null;
-    }
-
-    const tokens = this.props.workspace.semanticTokens(selectedHandle);
+    const tokens = this.props.workspace.semanticTokens(fileHandle);
     return generateMonacoTokens(tokens, model);
   }
 
@@ -245,25 +280,16 @@ class PlaygroundServer
     model: editor.ITextModel,
     range: Range,
   ): languages.SemanticTokens | null {
-    const selectedFile = this.props.files.selected;
-
-    if (selectedFile == null) {
-      return null;
-    }
-
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
       return null;
     }
 
     const tyRange = monacoRangeToTyRange(range);
-
     const tokens = this.props.workspace.semanticTokensInRange(
-      selectedHandle,
+      fileHandle,
       tyRange,
     );
-
     return generateMonacoTokens(tokens, model);
   }
 
@@ -271,20 +297,13 @@ class PlaygroundServer
     model: editor.ITextModel,
     position: Position,
   ): languages.ProviderResult<languages.CompletionList> {
-    const selectedFile = this.props.files.selected;
-
-    if (selectedFile == null) {
-      return;
-    }
-
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
-      return;
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
+      return undefined;
     }
 
     const completions = this.props.workspace.completions(
-      selectedHandle,
+      fileHandle,
       new TyPosition(position.lineNumber, position.column),
     );
 
@@ -300,6 +319,7 @@ class PlaygroundServer
             ? CompletionItemKind.Variable
             : mapCompletionKind(completion.kind),
         insertText: completion.name,
+        documentation: completion.documentation,
         // TODO(micha): It's unclear why this field is required for monaco but not VS Code.
         //  and omitting it works just fine? The LSP doesn't expose this information right now
         //  which is why we go with undefined for now.
@@ -318,20 +338,13 @@ class PlaygroundServer
     // eslint-disable-next-line @typescript-eslint/no-unused-vars
     _context: languages.SignatureHelpContext,
   ): languages.ProviderResult<languages.SignatureHelpResult> {
-    const selectedFile = this.props.files.selected;
-
-    if (selectedFile == null) {
-      return;
-    }
-
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
-      return;
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
+      return undefined;
     }
 
     const signatureHelp = this.props.workspace.signatureHelp(
-      selectedHandle,
+      fileHandle,
       new TyPosition(position.lineNumber, position.column),
     );
 
@@ -339,38 +352,15 @@ class PlaygroundServer
       return undefined;
     }
 
-    return {
-      dispose() {},
-      value: {
-        signatures: signatureHelp.signatures.map((sig) => ({
-          label: sig.label,
-          documentation: sig.documentation
-            ? { value: sig.documentation }
-            : undefined,
-          parameters: sig.parameters.map((param) => ({
-            label: param.label,
-            documentation: param.documentation
-              ? { value: param.documentation }
-              : undefined,
-          })),
-          activeParameter: sig.active_parameter,
-        })),
-        activeSignature: signatureHelp.active_signature ?? 0,
-        activeParameter:
-          signatureHelp.active_signature != null
-            ? (signatureHelp.signatures[signatureHelp.active_signature]
-                ?.active_parameter ?? 0)
-            : 0,
-      },
-    };
+    return this.formatSignatureHelp(signatureHelp);
   }
 
-  provideInlayHints(
-    _model: editor.ITextModel,
-    range: Range,
+  provideDocumentHighlights(
+    model: editor.ITextModel,
+    position: Position,
     // eslint-disable-next-line @typescript-eslint/no-unused-vars
     _token: CancellationToken,
-  ): languages.ProviderResult<languages.InlayHintList> {
+  ): languages.ProviderResult<languages.DocumentHighlight[]> {
     const workspace = this.props.workspace;
     const selectedFile = this.props.files.selected;
 
@@ -384,8 +374,30 @@ class PlaygroundServer
       return;
     }
 
-    const inlayHints = workspace.inlayHints(
+    const highlights = workspace.documentHighlights(
       selectedHandle,
+      new TyPosition(position.lineNumber, position.column),
+    );
+
+    return highlights.map((highlight: DocumentHighlight) => ({
+      range: tyRangeToMonacoRange(highlight.range),
+      kind: mapDocumentHighlightKind(highlight.kind),
+    }));
+  }
+
+  provideInlayHints(
+    model: editor.ITextModel,
+    range: Range,
+    // eslint-disable-next-line @typescript-eslint/no-unused-vars
+    _token: CancellationToken,
+  ): languages.ProviderResult<languages.InlayHintList> {
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
+      return undefined;
+    }
+
+    const inlayHints = this.props.workspace.inlayHints(
+      fileHandle,
       monacoRangeToTyRange(range),
     );
 
@@ -418,6 +430,66 @@ class PlaygroundServer
     this.props = props;
   }
 
+  private getOrCreateVendoredFileHandle(vendoredPath: string): FileHandle {
+    const cachedHandle = this.vendoredFileHandles.get(vendoredPath);
+    // Check if we already have a handle for this vendored file
+    if (cachedHandle != null) {
+      return cachedHandle;
+    }
+
+    // Use the new WASM method to get a proper file handle for the vendored file
+    const handle = this.props.workspace.getVendoredFile(vendoredPath);
+    this.vendoredFileHandles.set(vendoredPath, handle);
+    return handle;
+  }
+
+  private getFileHandleForModel(model: editor.ITextModel) {
+    // Handle vendored files
+    if (model.uri.scheme === "vendored") {
+      const vendoredPath = this.getVendoredPath(model.uri);
+
+      // If not cached, try to create it
+      return this.getOrCreateVendoredFileHandle(vendoredPath);
+    }
+
+    // Handle regular user files
+    const selectedFile = this.props.files.selected;
+    if (selectedFile == null) {
+      return null;
+    }
+
+    return this.props.files.handles[selectedFile];
+  }
+
+  private formatSignatureHelp(
+    signatureHelp: any,
+  ): languages.SignatureHelpResult {
+    return {
+      dispose() {},
+      value: {
+        signatures: signatureHelp.signatures.map((sig: any) => ({
+          label: sig.label,
+          documentation: sig.documentation
+            ? { value: sig.documentation }
+            : undefined,
+          parameters: sig.parameters.map((param: any) => ({
+            label: param.label,
+            documentation: param.documentation
+              ? { value: param.documentation }
+              : undefined,
+          })),
+          activeParameter: sig.active_parameter,
+        })),
+        activeSignature: signatureHelp.active_signature ?? 0,
+        activeParameter:
+          signatureHelp.active_signature != null
+            ? (signatureHelp.signatures[signatureHelp.active_signature]
+                ?.active_parameter ?? 0)
+            : 0,
+      },
+    };
+  }
+
   updateDiagnostics(diagnostics: Array<Diagnostic>) {
     if (this.props.files.selected == null) {
       return;
@@ -477,26 +549,18 @@ class PlaygroundServer
     // eslint-disable-next-line @typescript-eslint/no-unused-vars
     context?: languages.HoverContext<languages.Hover> | undefined,
   ): languages.ProviderResult<languages.Hover> {
-    const workspace = this.props.workspace;
-
-    const selectedFile = this.props.files.selected;
-    if (selectedFile == null) {
-      return;
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
+      return undefined;
     }
 
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
-      return;
-    }
-
-    const hover = workspace.hover(
-      selectedHandle,
+    const hover = this.props.workspace.hover(
+      fileHandle,
       new TyPosition(position.lineNumber, position.column),
     );
 
     if (hover == null) {
-      return;
+      return undefined;
     }
 
     return {
@@ -511,25 +575,17 @@ class PlaygroundServer
     // eslint-disable-next-line @typescript-eslint/no-unused-vars
     _: CancellationToken,
   ): languages.ProviderResult<languages.Definition | languages.LocationLink[]> {
-    const workspace = this.props.workspace;
-
-    const selectedFile = this.props.files.selected;
-    if (selectedFile == null) {
-      return;
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
+      return undefined;
     }
 
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
-      return;
-    }
-
-    const links = workspace.gotoTypeDefinition(
-      selectedHandle,
+    const links = this.props.workspace.gotoTypeDefinition(
+      fileHandle,
       new TyPosition(position.lineNumber, position.column),
     );
 
-    return mapNavigationTargets(links);
+    return this.mapNavigationTargets(links);
   }
 
   provideDeclaration(
@@ -538,25 +594,17 @@ class PlaygroundServer
     // eslint-disable-next-line @typescript-eslint/no-unused-vars
     _: CancellationToken,
   ): languages.ProviderResult<languages.Definition | languages.LocationLink[]> {
-    const workspace = this.props.workspace;
-
-    const selectedFile = this.props.files.selected;
-    if (selectedFile == null) {
-      return;
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
+      return undefined;
     }
 
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
-      return;
-    }
-
-    const links = workspace.gotoDeclaration(
-      selectedHandle,
+    const links = this.props.workspace.gotoDeclaration(
+      fileHandle,
       new TyPosition(position.lineNumber, position.column),
     );
 
-    return mapNavigationTargets(links);
+    return this.mapNavigationTargets(links);
   }
 
   provideDefinition(
@@ -565,25 +613,17 @@ class PlaygroundServer
     // eslint-disable-next-line @typescript-eslint/no-unused-vars
     _: CancellationToken,
   ): languages.ProviderResult<languages.Definition | languages.LocationLink[]> {
-    const workspace = this.props.workspace;
-
-    const selectedFile = this.props.files.selected;
-    if (selectedFile == null) {
-      return;
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
+      return undefined;
     }
 
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
-      return;
-    }
-
-    const links = workspace.gotoDefinition(
-      selectedHandle,
+    const links = this.props.workspace.gotoDefinition(
+      fileHandle,
       new TyPosition(position.lineNumber, position.column),
     );
 
-    return mapNavigationTargets(links);
+    return this.mapNavigationTargets(links);
   }
 
   provideReferences(
@@ -594,25 +634,17 @@ class PlaygroundServer
     // eslint-disable-next-line @typescript-eslint/no-unused-vars
     _: CancellationToken,
   ): languages.ProviderResult<languages.Location[]> {
-    const workspace = this.props.workspace;
-
-    const selectedFile = this.props.files.selected;
-    if (selectedFile == null) {
-      return;
+    const fileHandle = this.getFileHandleForModel(model);
+    if (fileHandle == null) {
+      return undefined;
     }
 
-    const selectedHandle = this.props.files.handles[selectedFile];
-
-    if (selectedHandle == null) {
-      return;
-    }
-
-    const links = workspace.gotoReferences(
-      selectedHandle,
+    const links = this.props.workspace.gotoReferences(
+      fileHandle,
       new TyPosition(position.lineNumber, position.column),
     );
 
-    return mapNavigationTargets(links);
+    return this.mapNavigationTargets(links);
   }
 
   openCodeEditor(
@@ -622,46 +654,44 @@ class PlaygroundServer
   ): boolean {
     const files = this.props.files;
 
-    const fileId = files.index.find((file) => {
-      return Uri.file(file.name).toString() === resource.toString();
-    })?.id;
-
-    if (fileId == null) {
+    // Model should already exist from mapNavigationTargets for both vendored and regular files
+    const model = this.monaco.editor.getModel(resource);
+    if (model == null) {
+      // Model should have been created by mapNavigationTargets
       return false;
     }
 
-    const handle = files.handles[fileId];
+    // Handle file-specific logic
+    if (resource.scheme === "vendored") {
+      // Get the file handle to track that we're viewing a vendored file
+      const vendoredPath = this.getVendoredPath(resource);
+      const fileHandle = this.getOrCreateVendoredFileHandle(vendoredPath);
+      this.props.onVendoredFileChange(fileHandle);
+    } else {
+      // Handle regular files
+      const fileId = files.index.find((file) => {
+        return Uri.file(file.name).toString() === resource.toString();
+      })?.id;
 
-    let model = this.monaco.editor.getModel(resource);
-    if (model == null) {
-      const language =
-        handle != null && isPythonFile(handle) ? "python" : undefined;
-      model = this.monaco.editor.createModel(
-        files.contents[fileId],
-        language,
-        resource,
-      );
+      if (fileId == null) {
+        return false;
+      }
+
+      // Set the model and trigger UI updates
+      if (files.selected !== fileId) {
+        this.props.onOpenFile(fileId);
+      }
     }
 
-    // it's a bit hacky to create the model manually
-    // but only using `onOpenFile` isn't enough
-    // because the model doesn't get updated until the next render.
-    if (files.selected !== fileId) {
-      source.setModel(model);
-
-      this.props.onOpenFile(fileId);
-    }
+    source.setModel(model);
 
     if (selectionOrPosition != null) {
       if (Position.isIPosition(selectionOrPosition)) {
         source.setPosition(selectionOrPosition);
-        source.revealPosition(selectionOrPosition);
+        source.revealPositionInCenterIfOutsideViewport(selectionOrPosition);
       } else {
         source.setSelection(selectionOrPosition);
-        source.revealPosition({
-          lineNumber: selectionOrPosition.startLineNumber,
-          column: selectionOrPosition.startColumn,
-        });
+        source.revealRangeNearTopIfOutsideViewport(selectionOrPosition);
       }
     }
 
@@ -694,6 +724,59 @@ class PlaygroundServer
     return null;
   }
 
+  private mapNavigationTargets(links: any[]): languages.LocationLink[] {
+    const result = links.map((link) => {
+      const uri = Uri.parse(link.path);
+
+      // Pre-create models to ensure peek definition works
+      if (this.monaco.editor.getModel(uri) == null) {
+        if (uri.scheme === "vendored") {
+          // Handle vendored files
+          const vendoredPath = this.getVendoredPath(uri);
+          const fileHandle = this.getOrCreateVendoredFileHandle(vendoredPath);
+          const content = this.props.workspace.sourceText(fileHandle);
+          this.monaco.editor.createModel(content, "python", uri);
+        } else {
+          // Handle regular files
+          const fileId = this.props.files.index.find((file) => {
+            return Uri.file(file.name).toString() === uri.toString();
+          })?.id;
+
+          if (fileId != null) {
+            const handle = this.props.files.handles[fileId];
+            if (handle != null) {
+              const language = isPythonFile(handle) ? "python" : undefined;
+              this.monaco.editor.createModel(
+                this.props.files.contents[fileId],
+                language,
+                uri,
+              );
+            }
+          }
+        }
+      }
+
+      const targetSelection =
+        link.selection_range == null
+          ? undefined
+          : tyRangeToMonacoRange(link.selection_range);
+
+      const originSelection =
+        link.origin_selection_range == null
+          ? undefined
+          : tyRangeToMonacoRange(link.origin_selection_range);
+
+      return {
+        uri: uri,
+        range: tyRangeToMonacoRange(link.full_range),
+        targetSelectionRange: targetSelection,
+        originSelectionRange: originSelection,
+      } as languages.LocationLink;
+    });
+
+    return result;
+  }
+
   dispose() {
     this.hoverDisposable.dispose();
     this.editorOpenerDisposable.dispose();
@@ -707,6 +790,7 @@ class PlaygroundServer
     this.semanticTokensDisposable.dispose();
     this.completionDisposable.dispose();
     this.signatureHelpDisposable.dispose();
+    this.documentHighlightDisposable.dispose();
   }
 }
 
@@ -759,29 +843,6 @@ function generateMonacoTokens(
   return { data: Uint32Array.from(result) };
 }
 
-function mapNavigationTargets(links: any[]): languages.LocationLink[] {
-  return links
-    .map((link) => {
-      const targetSelection =
-        link.selection_range == null
-          ? undefined
-          : tyRangeToMonacoRange(link.selection_range);
-
-      const originSelection =
-        link.origin_selection_range == null
-          ? undefined
-          : tyRangeToMonacoRange(link.origin_selection_range);
-
-      return {
-        uri: Uri.parse(link.path),
-        range: tyRangeToMonacoRange(link.full_range),
-        targetSelectionRange: targetSelection,
-        originSelectionRange: originSelection,
-      } as languages.LocationLink;
-    })
-    .filter((link) => link.uri.scheme !== "vendored");
-}
-
 function mapCompletionKind(kind: CompletionKind): CompletionItemKind {
   switch (kind) {
     case CompletionKind.Text:
@@ -836,3 +897,18 @@ function mapCompletionKind(kind: CompletionKind): CompletionItemKind {
       return CompletionItemKind.TypeParameter;
   }
 }
+
+function mapDocumentHighlightKind(
+  kind: DocumentHighlightKind,
+): languages.DocumentHighlightKind {
+  switch (kind) {
+    case DocumentHighlightKind.Text:
+      return languages.DocumentHighlightKind.Text;
+    case DocumentHighlightKind.Read:
+      return languages.DocumentHighlightKind.Read;
+    case DocumentHighlightKind.Write:
+      return languages.DocumentHighlightKind.Write;
+    default:
+      return languages.DocumentHighlightKind.Text;
+  }
+}
